import torch
from typing import List
from torch._dynamo.utils import CompileProfiler


__all__ = ["my_compiler","jit_backend","compare_prof"]

def my_compiler(gm: torch.fx.GraphModule, example_inputs: List[torch.Tensor],
                full_graph: bool = False,backend: str = "inductor",
                options: dict ={'trace.graph_diagram': True,'trace.enabled':True}):
    """
    This is a custom compiler for FX graphs. It is called by the
    `torch.fx.GraphModule` class when the `compile` method is called.
    """
    print("my_compiler() called with FX graph:")
    gm.graph.print_tabular()
    return gm.forward

def jit_backend(gm: torch.fx.GraphModule, example_inputs: List[torch.Tensor]):
    scripted = torch.jit.script(gm)
    return torch.jit.jit_backend(scripted)

def compare_prof(gm: torch.fx.GraphModule, example_inputs: List[torch.Tensor],logger=None):
    prof= CompileProfiler()
    prof_mode = torch.compile(gm,backend=prof)
    prof_mode(example_inputs)
    if logger:
        logger.info(prof.report())
    print(prof.report())


from mb_utils.src.logging import logger
from torch import nn
import torch
from torch.nn import functional as F
import torchvision
import torchvision.models as models
import os
import importlib

__all__ = ['ModelLoader','ModelExtractor']


def get_custom_model(data):
    """
    Function to get custom model from the models folder
    """
    model_name = data['model_name']
    model_custom = data['model_custom']
    model_module = importlib.import_module(model_custom)
    if model_name=='Unet':
        if data['unet_parameters']['attention']:
            model_name = 'Unet_attention'
        model_out = getattr(model_module, model_name)(**data['unet_parameters'])
    else:
        model_out = getattr(model_module, model_name)(**data['model_custom_params'])
    return model_out
    

class ModelLoader(nn.Module):
    def __init__(self, data : dict,logger=None):
        super().__init__()
        self._data= data 
        self._model_name=self._data['model_name']
        self._model_path=self._data['model_path']
        self._model_pretrained=self._data['model_pretrained']
        self._load_model = self._data['load_model']
        self._model_num_classes = self._data['model_num_classes']
        self._model_type=self._data['model_type']

    def model_type(self):
        """
        Function to get default model resnet, vgg, densenet, googlenet, inception, mobilenet, mnasnet, shufflenet_v2, squeezenet, or object_detection
        """

        if self._model_type=='detection':
            model_out = getattr(torchvision.models.detection,self._model_name)(pretrained=self._model_pretrained)
            return model_out


        if self._model_type=='classification':
            model_out = getattr(torchvision.models,self._model_name)(pretrained=self._model_pretrained)
            if hasattr(model_out,'fc'):
                num_ftrs = model_out.fc.in_features
                model_out.fc = nn.Linear(num_ftrs, self._model_num_classes)            
            if hasattr(model_out,'classifier'):
                for module in list(model_out.modules()):
                    if isinstance(module, nn.Linear):
                        first_layer = module
                        num_ftrs = first_layer.in_features
                        model_out.classifier = nn.Linear(num_ftrs, self._model_num_classes)
                        break
            return model_out
    

    def model_params(self):
        """
        Function to pass the model params to custom model
        """        
        #check if model is available in the models list
        model_out = get_custom_model(self._data)
        return model_out
        

    def get_model(self):
        """
        FUnction to get the model
        """
        # Check if the model is available in torchvision models

        if self._load_model:
            self.model = torch.load(self._data['load_model'])
            return self.model

        try:
            # Try to load the model from the specified path
            if hasattr(models, self._model_name) or hasattr(torchvision.models.detection, self._model_name):
                self.model = self.model_type() 
                if logger:
                    logger.info(f"Model {self._model_name} loaded from torchvision.models.") 
                return self.model
            else:
                self.model = self.model_params()
                return self.model
        except FileNotFoundError:
            raise ValueError(f"Model {self._model_name} not found in torchvision.models.")
    
    def forward(self,x):
        return self.model(x)
    

class ModelExtractor(nn.Module):
    def __init__(self, model):
        super(ModelExtractor, self).__init__()
        self.model = model
        self.feature_extractor = torch.nn.Sequential(*list(self.model.children())[:-1])
                
    def __call__(self, x):
        return self.feature_extractor(x)[:, :, 0, 0]


from torch import nn
import torch
from torch.nn import functional as F
import torchvision
import torchvision.models as models
from mb_pytorch.models.blocks.attention_block import Attention
from mb_pytorch.models.blocks.conv_block import Conv

__all__ = ['ModelOut']

class ModelOut(nn.Modules):
    def __init__(self, yaml_data,logger=None):
        self._data=yaml_data
        self._use_torchvision_models=self._data['use_torchvision_models']
        self._model_name=self._data['model_name']
        self._model_version=self._data['model_version']
        self._model_path=self._data['model_path']

    def model_type(model_name,model_version,pretrained=False):
        """
        Function to get default model resnet, vgg, densenet, googlenet, inception, mobilenet, mnasnet, shufflenet_v2, squeezenet
        """
        model_final = model_name+model_version
        model_out = getattr(models.model_name,model_final)(pretrained=pretrained)
        return model_out

    def get_model(self, model_name, model_version, model_path=None):
        """
        Function to get the model architecture from torchvision.models or from a specified path
        Inputs:
            model_name (str): Name of the model architecture
            model_path (str): Path to the model architecture
        Returns:
            model (torch.nn.Module): Model architecture
        """
        # Check if the model is available in torchvision models

        if self._use_torchvision_models==False:
            try:
                # Try to load the model from the specified path
                self.model = model_name
            except FileNotFoundError:
                raise ValueError(f"Model {model_name} not found in torchvision.models and path {model_path} does not exist.")

        else:
            if hasattr(models, model_name):
                model_class = getattr(models, model_name)
                if model_name in ['resnet', 'vgg', 'densenet', 'googlenet', 'inception', 'mobilenet', 'mnasnet', 'shufflenet_v2', 'squeezenet']:
                    # These models have pretrained weights available
                    self.model = self.model_type(model_name,model_version,pretrained=True)    

    
    def forward(self,x):
        return self.model(x)